import mcddns.provider as provider


class CheckIP(provider.AddressProvider):

    @classmethod
    def fetch(cls, options):
        return cls.fetch_url(
            "https://domains.google.com/checkip",
            timeout=options.timeout
        )



import abc
import importlib
import inspect
import ipaddress
import logging
import os
import pkgutil
import re
import requests
import sys
import mcddns.exceptions as exceptions
import mcddns.meta as meta


PROVIDER_MODULE_BUILTIN_PATH = os.path.join(
    os.path.dirname(__file__),
    "providers"
)
PROVIDER_MODULE_PREFIX = f"{meta.NAME}_"

log = logging.getLogger(__name__)


class Provider(abc.ABC):

    log = logging.getLogger(__name__)

    @classmethod
    def options_pre(cls, parser):
        pass

    @classmethod
    def options_post(cls, parser, options):
        pass


class AddressProvider(Provider):

    @classmethod
    @abc.abstractmethod
    def fetch(cls, options):
        pass

    @classmethod
    def validate(cls, options, address):
        return cls.is_ipv4_address(address)

    @staticmethod
    def fetch_url(url, timeout=3):
        try:
            return requests.get(url, timeout=timeout).text.strip()
        except requests.exceptions.RequestException as e:
            log.error(e)
            return None

    @staticmethod
    def is_ip_address(value):
        value = str(value)
        try:
            ipaddress.ip_address(value)
            log.debug("Valid IP address: %s", value)
            return True
        except ValueError:
            log.error("Invalid IP address: %s", value)
            return False

    @staticmethod
    def is_ipv4_address(value):
        value = str(value)
        try:
            if ipaddress.ip_address(value).version == 4:
                log.debug("Valid IPv4 address: %s", value)
                return True
            else:
                return False
        except ValueError:
            log.error("Invalid IPv4 address: %s", value)
            return False

    @staticmethod
    def is_ipv6_address(value):
        value = str(value)
        try:
            if ipaddress.ip_address(value).version == 6:
                log.debug("Valid IPv6 address: %s", value)
                return True
            else:
                return False
        except ValueError:
            log.error("Invalid IPv6 address: %s", value)
            return False

    @staticmethod
    def is_hostname(value):
        if len(value) > 255:
            result = False
        else:
            if value[-1] == ".":
                value = value[:-1]
            allowed = re.compile(r"(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
            result = all(allowed.match(x) for x in value.split("."))

        if result:
            log.debug("Valid hostname: %s", value)
        else:
            log.error("Invalid hostname: %s", value)
        return result


class DNSProvider(Provider):

    @classmethod
    @abc.abstractmethod
    def check(cls, options, address):
        pass

    @classmethod
    @abc.abstractmethod
    def update(cls, options, address):
        pass


def import_modules(*paths):
    sys.path = list(paths) + sys.path
    modules = [
        m.name for m in pkgutil.iter_modules()
        if m.name.startswith(PROVIDER_MODULE_PREFIX)
    ]
    return {m: importlib.import_module(m) for m in modules}


def strip_prefix(value, prefix):
    return (
        value[len(prefix):]
        if prefix and value.startswith(prefix)
        else value
    )


def provider_full_name(obj):
    module_name = strip_prefix(
        inspect.getmodule(obj).__name__,
        PROVIDER_MODULE_PREFIX
    )
    return f"{module_name}.{obj.__name__}"


def list_providers(*types):
    valid_types = Provider.__subclasses__()

    if not [t for t in types if t]:
        types = valid_types
    else:
        invalid_types = [str(t) for t in types if t not in valid_types]
        if invalid_types:
            raise exceptions.NoSuchProviderTypeError(
                f"Invalid provider types: {', '.join(invalid_types)}"
            )

    return {
        provider_full_name(cls): cls for cls
        in sum([t.__subclasses__() for t in types], [])
    }


def get_provider(name):
    try:
        return list_providers()[name]
    except KeyError:
        raise exceptions.NoSuchProviderError(f"No such provider: {name}")



import re
import mcddns.provider as provider


class CheckIP(provider.AddressProvider):

    @classmethod
    def fetch(cls, options):
        html = cls.fetch_url(
            "http://checkip.dyndns.com/",
            timeout=options.timeout
        )
        pattern = r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}"
        matches = re.search(pattern, str(html), re.ASCII)
        return matches[0] if matches else None



import boto3
import botocore.client
import logging
import mcddns.provider as provider


class CheckIP(provider.AddressProvider):

    @classmethod
    def fetch(cls, options):
        return cls.fetch_url(
            "https://checkip.amazonaws.com/",
            timeout=options.timeout
        )


class Route53(provider.DNSProvider):

    @classmethod
    def options_pre(cls, parser):
        parser.add_argument(
            "hosted_zone_id",
            help="aws route53 hosted zone id"
        )
        parser.add_argument(
            "--boto-log",
            action="store_true",
            help="show log messages from boto"
        )

    @classmethod
    def options_post(cls, parser, options):
        logging.getLogger("botocore").propagate = options.boto_log

    @classmethod
    def check(cls, options, address):
        if options.fqdn.endswith("."):
            fqdn = options.fqdn
        else:
            fqdn = f"{options.fqdn}."
        records = [{"Value": address}]

        cls.changes = []

        config = botocore.client.Config(connect_timeout=options.timeout,
                                        retries={'max_attempts': 0})
        cls.client = boto3.client("route53", config=config)

        rrsets = cls.client.list_resource_record_sets(
            HostedZoneId=options.hosted_zone_id,
            StartRecordName=fqdn,
            MaxItems="1",
        )["ResourceRecordSets"]

        if rrsets and rrsets[0]["Name"] == fqdn:
            cur_name = rrsets[0]["Name"].rstrip(".")
            cur_ttl = rrsets[0]["TTL"]
            cur_type = rrsets[0]["Type"]
            cur_records = rrsets[0]["ResourceRecords"]
            cur_address = " ".join(r["Value"] for r in cur_records)
            cur_record = f"{cur_name} {cur_ttl} {cur_type} {cur_address}"
            cls.log.debug("Current DNS record: %s", cur_record)

            if cur_type != options.rrtype:
                cls.changes.append({
                    "Action": "DELETE",
                    "ResourceRecordSet": rrsets[0]
                })
            elif cur_ttl == options.ttl and cur_records == records:
                return False
        else:
            cls.log.debug("DNS record not found: %s", options.fqdn)

        cls.changes.append({
            "Action": "UPSERT",
            "ResourceRecordSet": {
                "Name": fqdn,
                "Type": options.rrtype,
                "TTL": options.ttl,
                "ResourceRecords": records,
            }
        })

        return True

    @classmethod
    def update(cls, options, address):
        cls.client.change_resource_record_sets(
            HostedZoneId=options.hosted_zone_id,
            ChangeBatch={"Changes": cls.changes},
        )

        return True



NAME = __package__
DESCRIPTION = "Industrial-strength dynamic DNS client"
VERSION = "1.0.0"
AUTHOR = "Mike Conigliaro"
CONTACT = "http://conigliaro.org"
COPYRIGHT = f"""Copyright (c) 2020 {AUTHOR}"""



import argparse
import logging
import sys
import mcddns.meta as meta
import mcddns.provider as provider


log = logging.getLogger(__name__)


def parse(default_address_providers=[], args=None):
    addr_providers = provider.list_providers(provider.AddressProvider)
    addr_provider_names = sorted(addr_providers.keys())
    dns_providers = provider.list_providers(provider.DNSProvider)
    dns_provider_names = sorted(dns_providers.keys())

    parser = argparse.ArgumentParser(
        prog=meta.NAME,
        description=meta.DESCRIPTION,
        epilog=f"{meta.COPYRIGHT} ({meta.CONTACT})",
        # FIXME: https://bugs.python.org/issue27927
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"{meta.NAME} {meta.VERSION}"
    )

    subparsers = parser.add_subparsers(help='dns providers')
    for dns_provider_name, dns_provider in dns_providers.items():
        subparser = subparsers.add_parser(
            dns_provider_name,
            help=f"use the {dns_provider_name} provider"
        )
        subparser.add_argument(
            "fqdn",
            help="fully-qualified domain name"
        )
        subparser.add_argument(
            "-a",
            "--address-providers",
            choices=addr_provider_names,
            default=[],
            action="append",
            help="provider(s) used to obtain an address"
        )
        subparser.add_argument(
            "--rrtype",
            default="A",
            help="record type"
        )
        subparser.add_argument(
            "--ttl",
            type=int,
            default=300,
            help="time to live"
        )
        subparser.add_argument(
            "--timeout",
            type=float,
            default=3,
            help="timeout for network requests"
        )
        subparser.add_argument(
            "--retry",
            type=int,
            default=2,
            help="number of times to retry failed providers"
        )
        subparser.add_argument(
            "--no-backoff",
            action="store_true",
            help="disable fibonacci backoff for failed providers"
        )
        subparser.add_argument(
            "--cron",
            action="store_true",
            help="use exit codes suitable for cron"
        )
        subparser.add_argument(
            "--dry-run",
            action="store_true",
            help="show what will happen without making changes"
        )
        subparser.add_argument(
            "-l",
            "--log-level",
            choices=("debug", "info", "warning", "error", "critical"),
            default="info",
            help="show messages of this level or higher"
        )

        subparser.set_defaults(dns_provider=dns_provider_name)
        dns_provider.options_pre(subparser)
        for addr_provider_name, addr_provider in addr_providers.items():
            addr_provider.options_pre(parser.add_argument_group(
                f"{addr_provider_name} arguments")
            )

    options = parser.parse_args(args)

    if not hasattr(options, "dns_provider"):
        parser.print_help()
        parser.exit(2)

    if not options.address_providers:
        options.address_providers = default_address_providers

    for name in options.address_providers + [options.dns_provider]:
        provider.get_provider(name)().options_post(parser, options)

    if options.dry_run:
        log_format = "[%(levelname)s] (DRY-RUN) %(message)s"
    else:
        log_format = "[%(levelname)s] %(message)s"
    log_level = getattr(logging, options.log_level.upper())
    logging.basicConfig(format=log_format, level=log_level)

    log.debug("Provider search paths: %s", ', '.join(sys.path))
    log.debug("Found address providers: %s", ', '.join(addr_provider_names))
    log.debug("Found DNS providers: %s", ', '.join(dns_provider_names))
    options_str = ', '.join(f'{k}={repr(v)}' for k, v in
                            sorted(vars(options).items()))
    log.debug("Options: %s", options_str)

    return options



import mcddns.provider as provider


class IPv4(provider.AddressProvider):

    @classmethod
    def fetch(cls, options):
        return cls.fetch_url(
            "https://api.ipify.org/",
            timeout=options.timeout
        )


class IPv6(provider.AddressProvider):

    @classmethod
    def fetch(cls, options):
        return cls.fetch_url(
            "https://api6.ipify.org/",
            timeout=options.timeout
        )

    @classmethod
    def validate(cls, options, address):
        return cls.is_ip_address(address)



import itertools
import logging
import os
import tempfile
import time
import mcddns.meta as meta
import mcddns.provider as provider

RETURN_CODE_DRY_RUN = 100
RETURN_CODE_DNS_NOOP = 101
RETURN_CODE_DNS_UPDATED = 201
RETURN_CODE_ADDRESS_ERROR = 300
RETURN_CODE_DNS_ERROR = 301

RETURN_CODES_SUCCESS = range(100, 300)

CRON_IGNORE_RETURN_CODE_CLASS_TRANSITIONS = (
    (None, 1),
    (1, 1),
    (2, 1),
    (3, 3)
)

EXIT_CODE_SUCCESS = 0
EXIT_CODE_ERROR = 1

STATE_PATH = os.path.join(tempfile.gettempdir(), f"{meta.NAME}.state")


log = logging.getLogger(__name__)


def main(opts):
    tries = opts.retry + 1

    addr_providers = {p: provider.get_provider(p)
                      for p in opts.address_providers}
    addr_providers_with_retry = iterate_with_retry(
         addr_providers.items(),
         tries=tries,
         no_backoff=opts.no_backoff
    )
    for provider_name, provider_cls in addr_providers_with_retry:
        log.info("Using address provider: %s", provider_name)
        try:
            address = provider_cls.fetch(opts)
            log.info("Got address: %s", address)
            if provider_cls.validate(opts, address):
                break
        except Exception as e:
            log.exception(e)
    else:
        addr_provider_names = ', '.join(addr_providers.keys())
        log.critical("All address providers failed: %s", addr_provider_names)
        return RETURN_CODE_ADDRESS_ERROR

    desired_record = f"{opts.fqdn} {opts.ttl} {opts.rrtype} {address}"

    dns_providers_with_retry = iterate_with_retry(
        [provider.get_provider(opts.dns_provider)],
        tries=tries,
        no_backoff=opts.no_backoff
    )
    for provider_cls in dns_providers_with_retry:
        log.info("Using DNS provider: %s", opts.dns_provider)
        try:
            if provider_cls.check(opts, address):
                if opts.dry_run:
                    log.info("DNS will be updated: %s", desired_record)
                    return RETURN_CODE_DRY_RUN
                else:
                    if provider_cls.update(opts, address):
                        log.info("DNS updated: %s", desired_record)
                        return RETURN_CODE_DNS_UPDATED
                    else:
                        log.error("DNS update failed")
            else:
                log.info("No DNS update required")
                return RETURN_CODE_DNS_NOOP
        except Exception as e:
            log.exception(e)
    else:
        log.critical("DNS provider failed: %s", opts.dns_provider)
        return RETURN_CODE_DNS_ERROR


def fibonacci(n):
    if n == 0:
        return 0, 1
    else:
        a, b = fibonacci(n // 2)
        c = a * (b * 2 - a)
        d = a * a + b * b
        return (c, d) if n % 2 == 0 else (d, c + d)


def iterate_with_retry(iterable, tries=3, no_backoff=False):
    chain = list(itertools.chain.from_iterable(
        itertools.repeat(iterable, tries))
    )
    retries = len(chain) - 1
    for i, obj in enumerate(chain):
        if i:
            if no_backoff:
                log.info("Retrying (%d/%d)", i, retries)
            else:
                delay = fibonacci(i)[1]
                log.info("Retrying (%d/%d) in %ds...", i, retries, delay)
                time.sleep(delay)
        yield obj


def state_exists(path=STATE_PATH):
    exists = os.path.isfile(path)
    if exists:
        log.debug("Found previous state: %s", path)
    else:
        log.debug("No previous state found: %s", path)
    return exists


def state_read(path=STATE_PATH):
    if state_exists(path):
        try:
            with open(path, 'r') as f:
                data = f.read()
                log.debug("Read state: %s", data)
            return data
        except OSError as e:
            log.error("Unable to read state: %s", e)
            return ""
    else:
        return ""


def state_write(data, path=STATE_PATH):
    try:
        with open(path, 'w') as f:
            log.debug("Writing state: %s", data)
            f.write(str(data))
        return True
    except OSError as e:
        log.error("Unable to write state: %s", e)
        return False


def state_remove(path=STATE_PATH):
    if state_exists(path):
        try:
            log.debug("Removing previous state: %s", path)
            os.remove(path)
        except OSError as e:
            log.error("Unable to remove state: %s", e)
            return False
    return True


def return_code_class(return_code):
    try:
        return int(str(return_code)[0])
    except IndexError:
        None


def return_code_class_transition(return_code1, return_code2):
    return (
        return_code_class(return_code1),
        return_code_class(return_code2)
    )


def exit_code(return_code, cron=False, state_path=STATE_PATH):
    if return_code in RETURN_CODES_SUCCESS:
        exit_code = EXIT_CODE_SUCCESS
    else:
        exit_code = EXIT_CODE_ERROR

    if cron:
        previous_return_code = state_read(state_path)
        if str(return_code) != previous_return_code:
            state_write(return_code, path=state_path)

        transition = return_code_class_transition(
            previous_return_code,
            return_code
        )
        if transition in CRON_IGNORE_RETURN_CODE_CLASS_TRANSITIONS:
            cron_exit_code = EXIT_CODE_SUCCESS
        else:
            cron_exit_code = EXIT_CODE_ERROR

        if cron_exit_code == exit_code:
            log.debug("Exit code: %d", exit_code)
        else:
            log.debug(
                "Rewriting exit code for cron: %d => %d",
                exit_code, cron_exit_code
            )
            exit_code = cron_exit_code

    return exit_code



#!/usr/bin/env python

import os
import setuptools
import mcddns.meta as meta

setuptools.setup(
    name=meta.NAME,
    version=meta.VERSION,
    author=meta.AUTHOR,
    author_email="mike@conigliaro.org",
    description=meta.DESCRIPTION,
    long_description=open(os.path.join(
        os.path.dirname(__file__),
        "README.md")).read(),
    long_description_content_type="text/markdown",
    url="https://github.com/mconigliaro/mcddns",
    packages=setuptools.find_packages(),
    scripts=["bin/mcddns"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: Name Service (DNS)",
        "Topic :: System :: Systems Administration"
    ],
    python_requires=">=3.6",
    install_requires=[
        "boto3 >=1.12, <2.0",
        "requests >=2.23, <3.0"
    ],
)
